
/* midi.pl */
 /*  Heikki Välisuo */
/* Poimitaan midi-kommennot "laarista" ja tehdään midi-muotoinen kappale headereineen kaikkineen */
notes_out(Midi):-
    setof([A,B,C,D],midi_cmd4(A,B,C,D),SSn),
    setof([A,B,C],midi_cmd(A,B,C),SSm),
    message(setof),
    setof([A,B,C,D,E,F,G],tempo(A,B,C,D,E,F,G),SStempo),
%   message(duplicates(SStempo,SS7)),
   duplicates(SStempo,SS7),
    length(SSn,N),length(SSm,Nm),
   length(SS7,N7),
    append(SSn,SSm,SS0),  append(SS0,SS7,SS1),
    sort(SS1,SS),
%    time_signature(LenTs,T_signature),
 %   key_signature(LenKs,Key_signature),
    is(NBytes, 5*N + 3*Nm + 8*N7),  %   
    is(N1,NBytes+4),
    track_header(N1,MTrk),!,
    message(bytes_out(Midi,MTrk)),
    bytes_out(Midi,MTrk),!,
%    bytes_out(Midi,T_signature),!,
 %   bytes_out(Midi,Key_signature),!,
    message(midi_out(Midi,0,SS)),
    midi_out(Midi,0,SS),
    end_of_track(EOT),!, bytes_out(Midi,EOT).    

mthd_header(MTrcks,PpQN,Type,[["M","T","h","d",0,0,0,6,00,Type,00,MTrcks,00,PpQN]]).

track_header(NBytes,[["M","T","r","k"]|TLen]):-
    tr_len(NBytes,TLen).
%    write('track length  '), write(NBytes),
%    write(' '),write(TLen),nl.

time_signature(7,[255,88,4,4,2,24,8]).
key_signature(5,[255,89,2,0,0]).

end_of_track([00,FF,X2F,00]):- is(FF,15*16+15), is(X2F,2*16+15).   /* 0 FF 2F 00 */

duplicates([],[]).
duplicates([X|[]],[X]).
duplicates([X1|[X2|Xn]],[X1|Yn]):-
    X1 = [T1|_],
    X2 = [T2|_],
     T1 =\= T2,
%    write(dupl(T1,T2)),
   duplicates([X2|Xn],Yn).
duplicates([X1|[X2|Xn]],[X1|Yn]):-
    X1 = [T1|_],
    X2 = [T1|_],
    duplicates(Xn,Yn).

to_bytes(Ms,[M,L,K]):-
    is(M, Ms // round(256**2)), is(M1, Ms rem round(256**2)),
    is(L, M1 // 256), is(K, M1 rem 256).

/* */

midi_out(Midi,Dt0,[[_,Cmd0,P0,V0,A,B,C]|[]]):- !,
    put_dt(Midi,Dt0), put_byte(Midi,Cmd0),
    put_byte(Midi,P0),put_byte(Midi,V0),
    put_byte(Midi,A),put_byte(Midi,B),
    put_byte(Midi,C).
midi_out(Midi,Dt0,[[_,Cmd0,P0,V0]|[]]):- !,
    put_dt(Midi,Dt0), put_byte(Midi,Cmd0),
    put_byte(Midi,P0),put_byte(Midi,V0).
midi_out(Midi,Dt0,[[_,Cmd0,P0]|[]]):- !,
    put_byte(Midi,Dt0), put_byte(Midi,Cmd0),
    put_byte(Midi,P0).

%midi_out(Midi,Dt0,[[T0,Cmd0,P0,V0,A,B,C,D]|[[T1,Cmd1,P1,V1,A1,B1,C1,D1]|Tail]]):- !,
midi_out(Midi,Dt0,[[T0,Cmd0,P0,V0,A,B,C]|[[T1|Next]|Tail]]):- !,
    put_dt(Midi,Dt0), put_byte(Midi,Cmd0),
    put_byte(Midi,P0),put_byte(Midi,V0),
    put_byte(Midi,A),put_byte(Midi,B),
    put_byte(Midi,C),
    is(Dt1,(T1 - T0)),    
    midi_out(Midi,Dt1,[[T1|Next]|Tail]).
midi_out(Midi,Dt0,[[T0,Cmd0,P0,V0]|[[T1|Next]|Tail]]):- !,
    put_dt(Midi,Dt0), put_byte(Midi,Cmd0),
    put_byte(Midi,P0),put_byte(Midi,V0),
    is(Dt1,(T1 - T0)),    
    midi_out(Midi,Dt1,[[T1|Next]|Tail]).
midi_out(Midi,Dt0,[[T0,Cmd0,P0]|[[T1|Next]|Tail]]):- !,
    put_byte(Midi,Dt0), put_byte(Midi,Cmd0),
    put_byte(Midi,P0),
    is(Dt1,(T1 - T0)),    
    midi_out(Midi,Dt1,[[T1|Next]|Tail]).

bytes_out(_,[]):-!.
bytes_out(Str,X):- atomic(X),!, 
%    format(' ~10r ~c ',[X,","]), 
    put_byte(Str,X).
bytes_out(Str,[H|T]):- !, 
    bytes_out(Str,H), bytes_out(Str,T).


put_dt(F,Dt):-
    is(L0, Dt // 128), is(K, Dt rem 128),
    bit_up(L0,L),
    check_dt(Dt,L,K),
    (L0 > 0 ->put_byte(F,L); true), 
    put_byte(F,K).

check_dt(_,L,K):- L < 256, K < 256, L > -1, K > -1, !.
check_dt(Dt,L,K):- format("Dt: ~d ~d ~d ~n",Dt,L,K), halt.

tr_len(Nb,[N,M,L,K]):-
    is(N0,Nb // round(128**3)),  is(X1, Nb rem round(128**3)),
    is(M0, X1 // round(128**2)), is(X2, X1 rem round(128**2)),
    is(L0, X2 // 128), is(K, X2 rem 128),
    bit_up(N0,N), bit_up(M0,M), bit_up(L0,L).

bit_up(X,Y):- 
    is(Y,X+128).
    
lim_p(P0,P):- is(P, min(125,max(0,P0))).
lim_v(V0,V):- is(V, min(125,max(0,V0))).

/* */

dump_notes(CH):-
    setof([T,Cmd,P,V],midi_cmd4(T,Cmd,P,V),SS),
    sort(SS,SSz),
    dump_nots(CH,SSz).
dump_nots([],_).
dump_nots([Ch1|Chn],SSz):-
    bandi_name(Bname),
    format_to_atom(Pfilname,"data/~a_~d~a",[Bname,Ch1,'.pitch']),
    format_to_atom(Vfilname,"data/~a_~d~a",[Bname,Ch1,'.volum']),
    open(Vfilname,write,Vfil),
    open(Pfilname,write,Pfil),
    file_notes(Ch1,Pfil,Vfil,SSz),
    close(Pfil),close(Vfil),
dump_nots(Chn,SSz).

file_notes(_,_,_,[]).
file_notes(_,_,_,[[T,_,_,_]|_]):- T > 2160,!.
file_notes(Ch,P_File,V_File,[[T,Cmd,P,V]|Tail]):-
    (Ch+143 #=# Cmd -> (plot(P_File,T,P),plot(V_File,T,V)) ; true),
    (Ch+127 #=# Cmd -> plot(P_File,T,P); true),
    file_notes(Ch,P_File,V_File,Tail).

plot(File,X,Y):-
    format(File,"~d  ~d ~n",[X,Y]).

/* ++++++++++++ */

new_tempo(Time):-
    gtempo(MicroSecPerQ),
    is(Sae,Time // (3*288+36+3*19)),
    is(Vaihe, Sae mod 8),
    vaihe(Vaihe,Low,High),
    random(Low,High,N),
%   format(' time: ~d  tahti: ~d  vaihe: ~d  tempo: ~d~n',[Time,Tahti,Vaihe,N]),
    d_tempo(N,C),
    is(Y,round(C*MicroSecPerQ)),
    to_bytes(Y,[M,L,K]),
%    format('tempo: ~d  ~d ',[T,Y]),
    is(FF,15*16+15), is(H51,5*16+1),
    assertz(tempo(Time,FF,H51,3,M,L,K)),!.

%vaihe(Vaihe,-1,5):- 2 =< Vaihe,Vaihe =< 5,!. 
vaihe(0,-4,-2).
vaihe(1,-2,0).
vaihe(2,1,3).
vaihe(3,3,5).
vaihe(4,3,5).
vaihe(5,1,3).
vaihe(6,-2,0).
vaihe(7,-4,-2).
vaihe(X,-4,5):- write(vaihe(X)),nl,halt.

d_tempo(-4,0.7).
d_tempo(-3,0.75).
d_tempo(-2,0.8).
d_tempo(-1,0.9).
d_tempo(0,1.0).
d_tempo(1,1.1).
d_tempo(2,1.2).
d_tempo(3,1.25).
d_tempo(4,1.3).
